@@ -1,6 +1,49 @@
-(* sample file indented as we want it to be *)
+(* Sample file indented as we want it to be.  -*- tuareg -*- *)
 
-;; (* http://mantis/view.php?id=4247 *)
+let server_comments request t =
+  let module M = N in
+  let class M = N in
+  let m M = N in
+  let module M = N in
+  let open Grep.Server in
+  let x = 5 in
+  let modue x y = 5 in
+  let open M in
+
+  t >>= Grep.server_comments
+    lazy
+    parser
+    every
+
+let qs1 = {| quoted string |}   (* (issue #24) *)
+let qs2 = {eof| other quoted string   |noteof}  |eof}
+    
+let x =
+  let open M in
+  let x = 5 in
+  x + x
+;;
+
+(* FIXME: MAJOR "function" sends SMIE into a loop (fine with "fun").
+   Use M-q to test. *)
+let () =
+  let z = function t -> a in
+  foo z
+
+let () =
+  foo(function t -> a)
+;;
+
+(* FIXME: MAJOR: M-q on the "(" raises 'Scan error: "Unbalanced
+   parentheses"'.  It is fine if both () are on the same line. *)
+let () =
+  begin
+    (begin
+      end)
+  end
+;;
+
+;; (* http://caml.inria.fr/mantis/view.php?id=4247 *)
 let x = {
   Foo.
   a = b;
@@ -15,12 +58,12 @@ let x = {
       };
 }
 
-;; (* http://mantis/view.php?id=4249 *)
+;; (* http://caml.inria.fr/mantis/view.php?id=4249 *)
 let x = { a = b;
           c = d;
         }
 
-;; (* http://mantis/view.php?id=4255 *)
+;; (* http://caml.inria.fr/mantis/view.php?id=4255 *)
 { foo: [ `Foo of int
        | `Bar of string ];
 }
@@ -37,23 +80,23 @@ let a = {
 }
 
 let a = { t with M.
-  foo = foo;
-  bar = bar;
-}
+          foo = foo;
+          bar = bar;
+        }
 
 let a = { t with
-  M.
-  foo = foo;
-  bar = bar;
-}
+          M.
+          foo = foo;
+          bar = bar;
+        }
 
 type t = [ `Foo of int
          | `Bar of string ]
 
 type t =
-  | A
-  | B
-  | C
+| A
+| B
+| C
 with sexp
 
 type t = | A
@@ -66,26 +109,20 @@ type t = [
 | `C
 ]
 
-type t = [                              (* comment *)
+type t = [                              (* Comment.  *)
 | `A
 | `B
 | `C
 ]
 
-type t =
-  [ `A
-  | `B
-  | `C
-  ]
-
 type t = a
 and x = b
 
 module M = struct
   type t =
-    | A
-    | B
-    | C
+  | A
+  | B
+  | C
   with sexp
 
   type s = [
@@ -95,21 +132,29 @@ module M = struct
   ]
 
   type u =
-    | D
-    | E
+  | D
+  | E
+  with sexp
+end
+
+module N =
+struct
+  type u =
+  | D
+  | E
   with sexp
 end
 
 type m =
-  | T
+| T
 with sexp
 
-;; (* http://mantis/view.php?id=4334 *)
+;; (* http://caml.inria.fr/mantis/view.php?id=4334 *)
 type foo =
-    a
-    -> b
-    -> c
-    -> d
+  a
+  -> b
+  -> c
+  -> d
 
 val f :
   a:a
@@ -117,19 +162,19 @@ val f :
   -> c:c
 
 type bar = a -> b
-    -> c -> d
-    -> e -> f
+  -> c -> d
+  -> e -> f
 
 type baz = a -> b ->
-    c -> d ->
-    e -> f
+  c -> d ->
+  e -> f
 
 val quux : a -> b ->
   c -> d ->
   e -> f
 
 type t : a:b -> c:d
-    -> e:f -> g
+  -> e:f -> g
 
 val f : a:b -> c:d
   -> e:f -> g
@@ -200,51 +245,51 @@ type t = {
 } with sexp_of
 
 type 'a v = id:O.t ->
-    ssss:Ssss.t ->
-    dddd:ddd.t ->
-    t:S_m.t ->
-    mmm:Safe_float.t ->
-    qqq:int ->
-    c:C.t ->
-    uuuu:string option ->
-    aaaaaa:Aaaaaa.t ->
-    a:A.t ->
-    rrrrr:Rrrrr.t ->
-    time:Time.t ->
-    typ:[ `L_p of Safe_float.t ] ->
-    bazonk:present option ->
-    o_p_e:O_m.t option ->
-    only_hjkl:present option ->
-    show_junk:int option ->
-    d_p_o: Safe_float.t option ->
-    asdf:present option ->
-    generic:Sexp.t list ->
-    'a
+  ssss:Ssss.t ->
+  dddd:ddd.t ->
+  t:S_m.t ->
+  mmm:Safe_float.t ->
+  qqq:int ->
+  c:C.t ->
+  uuuu:string option ->
+  aaaaaa:Aaaaaa.t ->
+  a:A.t ->
+  rrrrr:Rrrrr.t ->
+  time:Time.t ->
+  typ:[ `L_p of Safe_float.t ] ->
+  bazonk:present option ->
+  o_p_e:O_m.t option ->
+  only_hjkl:present option ->
+  show_junk:int option ->
+  d_p_o: Safe_float.t option ->
+  asdf:present option ->
+  generic:Sexp.t list ->
+  'a
 
 type 'a v =
-    id:O.t
-    -> ssss:Ssss.t
-    -> dddd:ddd.t
-    -> t:S_m.t
-    -> mmm:Safe_float.t
-    -> qqq:int
-    -> c:C.t
-    -> uuuu:string option
-    -> aaaaaa:Aaaaaa.t
-    -> a:A.t
-    -> rrrrr:Rrrrr.t
-    -> time:Time.t
-    -> typ:[ `L_p of Safe_float.t ]
-    -> bazonk:present option
-    -> o_p_e:O_m.t option
-    -> only_hjkl:present option
-    -> show_junk:int option
-    -> d_p_o: Safe_float.t option
-    -> asdf:present option
-    -> generic:Sexp.t list
-    -> 'a
+  id:O.t
+  -> ssss:Ssss.t
+  -> dddd:ddd.t
+  -> t:S_m.t
+  -> mmm:Safe_float.t
+  -> qqq:int
+  -> c:C.t
+  -> uuuu:string option
+  -> aaaaaa:Aaaaaa.t
+  -> a:A.t
+  -> rrrrr:Rrrrr.t
+  -> time:Time.t
+  -> typ:[ `L_p of Safe_float.t ]
+  -> bazonk:present option
+  -> o_p_e:O_m.t option
+  -> only_hjkl:present option
+  -> show_junk:int option
+  -> d_p_o: Safe_float.t option
+  -> asdf:present option
+  -> generic:Sexp.t list
+  -> 'a
 
-;; (* not in mantis *)
+;; (* Not in mantis.  *)
 let bar x =
   if y
   then x
@@ -252,12 +297,12 @@ let bar x =
 
 let zot x =
   quux ~f:(if x
-    then y
-    else z)
+           then y
+           else z)
 
 let zot x = quux ~f:(if x
-  then y
-  else z)
+                     then y
+                     else z)
 
 let () =
   if foo
@@ -286,17 +331,38 @@ let () =
   )
 
 let rec count_append l1 l2 count =
+  (* http://caml.inria.fr/resources/doc/guides/guidelines.en.html *)
   match l1 with
-    | []               ->                         l2
-    | [x1]             -> x1                   :: l2
-    | [x1; x2]         -> x1 :: x2             :: l2
-    | [x1; x2; x3]     -> x1 :: x2 :: x3       :: l2
-    | [x1; x2; x3; x4] -> x1 :: x2 :: x3 :: x4 :: l2
-    | x1 :: x2 :: x3 :: x4 :: x5 :: tl ->
-      x1 :: x2 :: x3 :: x4 :: x5 ::
-        (if count > 1000
-         then slow_append tl l2
-         else count_append tl l2 (count + 1))
+  | []               ->                         l2
+  | [x1]             -> x1                   :: l2
+  | [x1; x2]         -> x1 :: x2             :: l2
+  | [x1; x2; x3]     -> x1 :: x2 :: x3       :: l2
+  | [x1; x2; x3; x4] -> x1 :: x2 :: x3 :: x4 :: l2
+  | x1 :: x2 :: x3 :: x4 :: x5 :: tl ->
+     x1 :: x2 :: x3 :: x4 :: x5 ::
+       (if count > 1000
+        then slow_append tl l2
+        else count_append tl l2 (count + 1))
+
+
+let x =
+  match x with
+  | Foo of
+      < tag : t;  (* FIXME *)
+      md :  t;
+      is_me :  t;
+      >
+;;
+
+let x =
+  match x with
+  | Foo of
+      <
+        tag : t;  (* FIXME *)
+      md :  t;
+      is_me :  t;
+      >
+;;
 
 let foo =
   (
@@ -314,6 +380,7 @@ let foo x = function
   | None -> false
 
 let bar x = fun u ->
+  match u with
   | Some _ -> true
   | None -> false
 
@@ -327,8 +394,8 @@ let () = match x with
 
 let () =
   match x with
-      Foo -> 1
-    | Bar -> 2
+    Foo -> 1
+  | Bar -> 2
 
 let r x =
   try  f x;
@@ -338,39 +405,43 @@ let r x =
 
 let g x =
   try let a = b in
-        f x;
-        g x;
-        y x;
+      f x;
+      g x;
+      y x;
   with e -> raise e
 
 let h x =
   try  ff a b
-         c d;
+          c d;
        gg 1 2
-         3 4;
+          3 4;
   with e -> raise e
 
 let () =
   try
     _
   with
-      Bar -> ()
+    Bar -> ()
 
 let () =
+  (* http://caml.inria.fr/resources/doc/guides/guidelines.en.html *)
   try () with
-    |e ->
-      let x = z in
+  | e ->
+    let x = z in
 
-        yyyyy
-          (a b)
+    yyyyy
+      (a b)
 
 let d x = function
+  (* FIXME: Should we leave it like this or align "|" with "match"?
+     I chose with "match" because it looks otherwise odd and is more
+     consistent with the "try" alignments above.  *)
   | A -> (match x with
-      | X ->
-        false
-      | Y -> true
-      |  Z ->
-        false)
+         | X ->
+            false
+         | Y -> true
+         |  Z ->
+             false)
   | B -> false
 
 let a f = function
@@ -380,27 +451,27 @@ let a f = function
     2
   |      C ->
     (function
-      |  X  ->
-        a
-      | Y ->
-        b) 12
+    |  X  ->
+      a
+    | Y ->
+      b) 12
   | D ->
     (match z with
-      | 4 -> 3
-      |  5 -> 7)
+     | 4 -> 3
+     |  5 -> 7)
 
-let x = foo ~f:(fun _ -> 0              (* comment *)
+let x = foo ~f:(fun _ -> 0              (* Comment.  *)
 )
 
 let f x =
   (let y = x in
-     f x;
-     g y;
-     h z)
+   f x;
+   g y;
+   h z)
 
 let f x =
   (let y = x in
-     f x);
+   f x);
   g y;
   h z
 
@@ -408,15 +479,15 @@ let g y =
   a b;
   c d;
   e f;
-  (* comment *)
+  (* Comment.  *)
   g h;
   i j
 
 let () =
   (let a = 1 in
    let b = 2 in
-     ( a,
-       b))
+   ( a,
+     b))
 
 let () =
   ((a b
@@ -431,32 +502,32 @@ let () =
   if a
   then
     let b = P.s ~b ~a ~m in
-      a +. e *. b,
-      b -. e *. b
+    a +. e *. b,
+    b -. e *. b
   else
     q.a -. s *. z,
     q.b +. s *. z
 
 let () =
-  (* comment *)
+  (* Comment.  *)
   (let x =
      3
    in
-     x + 5)
+   x + 5)
 
 let foo = 1 and bar = 2 and zot = 3 in
 let quux = 4 in
-  foo
-  + bar
-  + zot
-  + quux
+foo
++ bar
++ zot
++ quux
 
-(* indent comment to following code *)
+(* Indent comment to following code.  *)
 let () =
   try                                   (* foo!
                                            bar *)
     let a = f g c d in
-      a b
+    a b
   with _ -> ()
 
 let () = try
@@ -477,14 +548,14 @@ let () =
   try f a
   with A () ->
     ()
-    | B () ->
-      ()
-    |     C () ->
-      ()
+  | B () ->
+    ()
+  |     C () ->
+    ()
 
 let f errors input =
   let ( @@ ) string bool = if not bool then errors := string :: !errors in
-    input @@ false
+  input @@ false
 
 let x =
   if mode = foo then bar;
@@ -495,16 +566,16 @@ let x =
 
 let () =
   match _ with
-    | foo ->
-      bar
-      >>| function _ ->
-        _
+  | foo ->
+     bar
+     >>| function _ ->
+                  _
 
 let () =
   a
   >>= fun () ->
   b
-  >>= fun () ->
+  >>| fun () ->
   Deferred.all
 
 let x =
@@ -522,8 +593,8 @@ let x =
     4 5 6 >>= fun y ->
   w*3 >>= fun q -> r
 
-(* this does not work, see comment in tuareg-compute-arrow-indent
- * workaround: wrap code in parens *)
+(* This does not work, see comment in tuareg-compute-arrow-indent.
+ * Workaround: wrap code in parens.  *)
 (* let () =
  *   match
  *     a 1 2 3
@@ -537,9 +608,9 @@ let () =
   match
     let a = a in
     let b = b in
-      c
+    c
   with
-    | A -> _
+  | A -> _
 
 let () =
   match
@@ -547,12 +618,12 @@ let () =
      b >>= fun b ->
      c)
   with
-      A -> _
+    A -> _
 
 let f t =
   let (a, b) = to_open in
   let c = g t a b in
-    ()
+  ()
 
 let () =
   begin
@@ -580,15 +651,15 @@ let () =
 
 let () =
   match e with
-    | `T d ->
-      notify `O `T d;
-      cancel t u ~now
+  | `T d ->
+    notify `O `T d;
+    cancel t u ~now
 
 let () =
   let a = 1
   and b = 2
   and c = 3 in
-    a + b + c
+  a + b + c
 
 let _ =
   foo bar
@@ -608,10 +679,10 @@ let () =
   Config.load ()
   >>> fun config ->
   let quux = config.Config.bazonk.Config.Bazonk.quux in
-    load_quux ~input quux config
+  load_quux ~input quux config
   >>> fun quux ->
   let da = Poo.Snapshot.merge quux in
-    load_foobar config ~input
+  load_foobar config ~input
   >>> fun foobar ->
   whatever foobar
 
@@ -623,10 +694,10 @@ let () =
 let () =
   a
   >>= function
-    | b -> c
-    | d ->
-      e
-      >>= f
+  | b -> c
+  | d ->
+    e
+    >>= f
 
 let () =
   foo >>> fun bar ->
@@ -637,7 +708,7 @@ let () =
   Config.load ()
   >>> fun config ->
   let quux = x in
-    x
+  x
   >>= fun quux ->
   x
 
@@ -645,7 +716,7 @@ let () =
   Config.load ()
   >>= fun config ->
   let quux = x in
-    x
+  x
   >>= fun quux ->
   x
 
@@ -666,21 +737,23 @@ let () =
 
 let () =
   (let a,b = match c with
-    | D -> e,f
-    | G -> h,i in
+     | D -> e,f
+     | G -> h,i in
    let j = a + b in
-     j * j),
+   j * j),
   12
 
 module type M = M2
   with type t1 = int
+   and type t2 = int
+   and module S = M3
   with type t2 = int
   with type t3 = int
 
 let () =
   try
     match () with
-      | () -> ()
+    | () -> ()
   with _ -> ()
 
 let () =
@@ -712,10 +785,10 @@ let x =
     (a
      ^ b
      ^ c) in
-    f ~a:b ?c:d
-      ?e:f ~g:(h i j)
-      ~k:(l m)
-      (n o p)
+  f ~a:b ?c:d
+    ?e:f ~g:(h i j)
+    ~k:(l m)
+    (n o p)
 
 let () =
   foobar (fun () ->
@@ -726,8 +799,8 @@ let w f =
   List.map f ~f:(fun (a, b) ->
     L.r a
     >>= function
-      | Ok s -> `Fst (b, s)
-      | Error e -> `Snd (b, a, e))
+    | Ok s -> `Fst (b, s)
+    | Error e -> `Snd (b, a, e))
 
 class c (a : b) =
 object
@@ -744,7 +817,7 @@ let f a = {
 }
 
 let f a
-    b = {
+      b = {
   a = a;
   b = b;
 }
@@ -756,19 +829,9 @@ let () =
 
 let a =
   B.c d ~e:f [
-    "g";
-    "h";
-  ]
-
-let a = match b with
-  | Some c -> Some {
-    d = c;
-    e = e
-  }
-  | None -> {
-    d = c;
-    e = e
-  }
+        "g";
+        "h";
+      ]
 
 let () =
   f a ~b:c ~d ~e:g
@@ -777,10 +840,20 @@ let () =
       "b";
     ]
 
+let a = match b with
+  | Some c ->  Some {
+    d = c;
+    e = e
+  }
+  | None -> {
+    d = c;
+    e = e
+  }
+
 let a = {
   b = (
     let z = f u in
-      z + z;
+    z + z;
   );
   c = (let a = b in {
     z = z;
@@ -794,8 +867,8 @@ let () =
       C.d e ~f:(fun g -> (h.I.j.K.l, m))
       |! begin fun n ->
         match O.p n with
-          | `Q r -> r
-          | `S _k -> assert false
+        | `Q r -> r
+        | `S _k -> assert false
       end;
     t =
       u ~v:w
@@ -853,7 +926,7 @@ let a =
 let a =
   foo
     ~f:(fun () ->
-      a)
+         a)
 
 let a =
   foo
@@ -863,21 +936,21 @@ let a =
 let a =
   foo
     ~f:(fun () -> a
-    )
+       )
 
 let () =
-  (* comment *)
+  (* Comment.  *)
   bar a b
     c d;
   foo ~size
-    (* comment *)
+    (* Comment.  *)
     ~min:foo
     ?reduce
     ?override
     ()
 
 let foo =
-  (* comment *)
+  (* Comment.  *)
   List.map z
     ~f:(fun m ->
       M.q m
@@ -910,22 +983,23 @@ let () =
 
 let variants a =
   match String.split a ~on:'-' with
-    | [ s1; s2; s3 ] ->
-      let a0 = String.concat ~sep:"" [ s1; s2] in
-      let a1 = String.concat ~sep:"-" [ s1; s2; s3; "055" ] in (* comment *)
-        List.map [ a0; a1; a]
-          ~f:(fun a_s -> lookup a_s)
-        |! List.flatten
-    | _ -> failwith "bad"
+  | [ s1; s2; s3 ] ->
+    let a0 = String.concat ~sep:"" [ s1; s2] in
+    let a1 = String.concat ~sep:"-" [ s1; s2; s3; "055" ] in (* Comment.  *)
+    List.map [ a0; a1; a]
+      ~f:(fun a_s -> lookup a_s)
+    |! List.flatten
+  | _ -> failwith "bad"
 
 let f a1 a2 a3
-    b1 b2 b3 d1 d2 d3 = {
+      b1 b2 b3 d1 d2 d3 = {
   aa = func1 a1 a2 a3;
   bb = func2
     b1 b2 b3;
+  (* FIXME: Here it is reasonable to have '|' aligned with 'match' *)
   cc = (match c with
-    | A -> 1
-    | B -> 2);
+        | A -> 1
+        | B -> 2);
   dd = func3
     d1 d2 d3;
 }
@@ -941,11 +1015,227 @@ let fv =
       z
     ->
       match x y z with
-        | `No)
+      | `No)
 
 (* https://forge.ocamlcore.org/tracker/index.php?func=detail&aid=644&group_id=43&atid=255 *)
 let b =
   match z with
-    | 0 -> fun x -> x
-    | 1 -> fun x -> 1
+  | 0 -> fun x -> x
+  | 1 -> fun x -> 1
 
+
+module type X =
+struct
+  val f : float -> float
+  (** This comment should be under "val", like other doc comments and not
+   aligned to the left margin. *)
+end
+
+let test () =                           (* bug#927 *)
+  if a then
+    if b then x
+    else if c then y
+    else z
+  else something
+
+let f x =
+  if x = 1 then print "hello";
+  print "there";
+  print "everywhere"
+
+let f x =
+  if print "hello"; x = 1 then print "hello";
+  print "there"
+
+let f x =
+  if x = 1 then let y = 2 in print "hello";
+                             print "there"
+  else print "toto"
+
+let f x =
+  match x with
+  | 1 -> let x = 2 in
+         if x = 1 then print "hello"
+  | 2 -> print "there"
+
+let f x =
+  if x = 1 then match x with
+                | 1 -> print "hello"
+                | 2 -> print "there"
+  else print "toto"
+
+let f x =
+  x + 4 +
+    x + 5 +
+    x + 6
+
+let splitting_long_expression =
+  quad.{band, i3} <- quad.{band, i3} +. g +.
+                       area_12 *. (P.potential x13 y13 +. P.potential x23 y23)
+
+let () =
+  (* Beware of lexing ".;" as a single token!  *)
+  A.Axes.box vp;
+  A.fx vp (E.on_ray u0) 0. 2000.;
+  A.Viewport.set_color vp A.Color.green
+
+let f x =
+  1
+and g y =
+  2
+
+let x =
+  let module M =
+    struct
+    end
+  in 0
+
+let x =
+  try a
+  with Not_found ->
+    b
+let x =
+  try a
+  with Not_found ->
+       b
+     | _ ->
+       c
+let x =
+  try a
+  with Not_found ->
+       if a then b
+     | flag when String.is_prefix flag ~prefix:"-" ->
+       a
+     | _ ->
+       c
+
+let x = "toto try \
+         tata"
+
+let optional_sci_float =
+  do_something ~a:1e-7
+               ~b:(fun x -> x + 1)
+
+let () =
+  f x ~tol:1.0
+    more arguments;
+  f x ~tol:1.
+    more arguments
+
+let array_args =
+  fold s multi_sms.(0).message_number folder
+       more_args (* FIXME *)
+
+let () =
+  match var with
+  | <:expr< $lid:f$ >> ->
+     KO
+  | <:expr< $lid:f$ >> when f x ->
+     KO
+  | y when f y ->
+     OK
+  | long_pattern
+       when f long_pattern -> (* Should be more indented than the clause body *)
+     z
+
+type t = {
+  mutable a: float;
+  b : int;
+}
+
+(* [struct] and [sig] must be treated the same way. *)
+module Base64 : sig
+  val f : int -> int
+end
+
+external f :
+  int -> unit  (* Treated as [val].  *)
+  = "f_stub"
+
+let () =
+  g a.[k]
+    x   (* aligned with [a], despite the dot *)
+
+let () =
+  g a.[k] 1.0
+    x   (* aligned with [a], despite the dots *)
+
+(* OOP elements (from Marc Simpson <marc AT 0branch DOT com>).  *)
+
+class useless = object
+  val n = 10
+
+  method incremented () =
+    succ n
+
+  method add_option = function
+    | Some x -> Some(n + x)
+    | None   -> None
+end
+
+class useless' = object(self)
+  val n = 10
+
+  method incremented () =
+    succ n
+
+  method add_option = function
+    | Some x -> Some(n + x)
+    | None   -> None
+end
+
+class useless' = object(self)
+  val n = 10
+
+  initializer
+    print_endline "Initialised."
+
+  method incremented () =
+    succ n
+
+  method private add x =
+    n + x
+
+  method add_option = function
+    | Some x -> Some(self#add x)
+    | None   -> None
+end
+
+(* Signatures with labeled arguments *)
+
+val f :
+  x : int ->
+  int ->
+  int
+
+val f :
+  ?x : int ->
+  int ->
+  int
+
+let subscribe_impl dir topic ~aborted =
+  return (
+      match Directory.subscribe dir topic with
+      | None -> Error ()
+      | Some pipe ->
+         whenever (aborted >>| fun () -> Pipe.close_read pipe);
+         Ok pipe
+    )
+    next_argument (* should be indented correctly, given the braces *)
+
+
+List.map
+  (function x ->
+    blabla    (* FIXME: indentation afer "(function" *)
+    blabla
+    blabla)
+  l
+
+let command =
+  Command.Spec.(
+    empty
+    +> flag "-hello" (optional_with_default "Hello" string)
+            ~doc:" The 'hello' of 'hello world'"
+    +> flag "-world" (optional_with_default "World" string)
+            ~doc:" The 'world' of 'hello world'"
+  )
